# Used to created a resource that replicates data
# between 2 hosts for HA.
#
# == Parameters
#  [host1] Name of first host. Required unless $cluster is set.
#  [host2] Name of second host. Required unless $cluster is set.
#  [ip1] Ipaddress of first host. Required unless $cluster or $res1/$res2 is set.
#  [ip2] Ipaddress of second host. Required unless $cluster or $res1/$res2 is set.
#  [res1] First stacked resource name.
#  [res2] Second stacked resource name.
#  [disk] Name of disk to be replicated. Assumes that the
#     name of the disk will be the same on both hosts. Required.
#  [secret] The shared secret used in peer authentication.. False indicates that
#    no secret be required. Optional. Defaults to false.
#  [port] Port which drbd will use for replication on both hosts.
#     Optional. Defaults to 7789.
#  [protocol] Protocol to use for drbd. Optional. Defaults to 'C'
#     http://www.drbd.org/users-guide/s-replication-protocols.html
#  [verify_alg] Algorithm used for block validation on peers. Optional.
#    Defaults to crc32c. Accepts crc32c, sha1, or md5.
#  [disk_parameters] Parameters for disk{} section
#  [manage] If the actual drbd resource shoudl be managed.
#  [ha_primary] If the resource is being applied on the primary host.
#  [initial_setup] If this run is associated with the initial setup. Allows a user
#    to only perform dangerous setup on the initial run.
#  [link_util] replication link network utilization percent
#  [link_speed] replication link network speed mbps
#  [num_parallel] number of parallel drbd filesystems to sync
#  [rtt_ms] round-trip-time milliseconds (i.e., ping between replication nodes)
#  [cpumask] cpu-affinity-mask for DRBD kernel threads (hexidecimal notation).
#    0 means spread over all CPUs of the machine.
define drbd::resource (
  $host1          = undef,
  $host2          = undef,
  $ip1            = undef,
  $ip2            = undef,
  $res1           = undef,
  $res2           = undef,
  $cluster        = undef,
  $secret         = false,
  $port           = '7789',
  $device         = '/dev/drbd0',
  $mountpoint     = "/drbd/${name}",
  $automount      = true,
  $owner          = 'root',
  $group          = 'root',
  $protocol       = 'C',
  $verify_alg     = 'crc32c',
  $disk_parameters = false,
  $link_util      = false,
  $link_speed     = false,
  $num_parallel   = false,
  $rtt_ms         = false,
  $resync_after   = undef,
  $cpumask        = false,
  $net_parameters = false,
  $manage         = true,
  $ha_primary     = false,
  $initial_setup  = false,
  $fs_type        = 'ext4',
  $mkfs_opts      = '',
  $disk           = undef,
  $handlers       = false,
) {
  include ::drbd

  Exec {
    path      => ['/bin', '/sbin', '/usr/bin'],
    logoutput => 'on_failure',
  }

  File {
    owner => 'root',
    group => 'root',
  }

  file { $mountpoint:
    ensure => directory,
    mode   => '0755',
    owner  => $owner,
    group  => $group,
  }

  if $link_util and $link_speed and $num_parallel and $rtt_ms {
    # Engineer drbd variable sync rate parameters based on the following:
    #  https://blogs.linbit.com/p/128/drbd-sync-rate-controller/
    #  https://blogs.linbit.com/p/443/drbd-sync-rate-controller-2/
    # Methodology adapted to account for replication link speed and parallelism.

    # Since there is no aggregate bandwidth control, prorate the drbd
    # replication bandwidth based on parallelism.
    # Based on experimentation, it seems generally better to set num_parallel
    # to 1 and let DRBD auto-regulate its throughput.  The end result is that
    # multiple competing filesystems (i.e., on same disk device) already have
    # their sync throughput reduced.
    $mbps = $link_speed / Numeric($num_parallel)

    # bandwidth delay product
    $bdp_k = $mbps * Numeric($rtt_ms)

    # engineer initial sync rate as percent of link bandwidth
    $rate_M = floor(Numeric($link_util) * $mbps / 8 / 100)
    $rate = "${rate_M}M"

    # engineer c_plan_ahead to default value (tenths)
    # Documentation indicates this value OK even for 200 ms RTT.
    $c_plan_ahead = 20

    # engineer c_fill_target as 1*BDP (tune within 1x to 3x BDP;
    # choose minimum value that saturates bandwidth)
    $fill_target_k = floor(1 * $bdp_k)
    $c_fill_target = "${fill_target_k}k"

    # engineer c_min_rate -- experimentally determined so DRBD is not
    # throttled to a crawl even when there is minimal application IO.
    # DRBD default is way too small.
    $min_rate_M = 15 + floor(Numeric($link_util) * $mbps / 8 / 100 / 25)
    $c_min_rate = "${min_rate_M}M"

    # engineer c_max_rate as percent of link bandwidth
    $max_rate_M = floor(Numeric($link_util) * $mbps / 8 / 100)
    $c_max_rate = "${max_rate_M}M"

    # various tuning settings to enable larger link bandwidth (eg, 10G)
    # max_buffers should scale with MBps; set to maximum settable
    $max_buffers = 20000
    $max_epoch_size = 20000
    $unplug_watermark = 16
    # sndbuf_size and rcvbuf_size should scale with mbps; set 0 to auto-tune
    $sndbuf_size = 0
    $rcvbuf_size = 0
    # increase al_extents to improve random write throughput; set to prime number
    $al_extents = 3389
  } else {
    # disable variable sync rate
    $c_plan_ahead  = 0
    $c_fill_target = false
    $c_min_rate    = false
    $c_max_rate    = false

    # engineer fixed sync rate at 40 percent of 1G
    $rate_M = floor(40 * 1000 / 8 / 100)
    $rate = "${rate_M}M"

    $max_buffers = false
    $max_epoch_size = false
    $unplug_watermark = false
    $sndbuf_size = false
    $rcvbuf_size = false
    $al_extents = false
  }

  concat { "/etc/drbd.d/${name}.res":
    mode    => '0600',
    require => [
      Package['drbd'],
      File['/etc/drbd.d'],
    ],
    notify  => Class['drbd::service'],
  }
  # Template uses:
  # - $name
  # - $protocol
  # - $device
  # - $disk
  # - $secret
  # - $verify_alg
  # - $host1
  # - $host2
  # - $ip1
  # - $ip2
  # - $port
  concat::fragment { "${name} drbd header":
    target  => "/etc/drbd.d/${name}.res",
    content => template('drbd/header.res.erb'),
    order   => '01',
  }
  # Export our fragment for the clustered node
  if $ha_primary and $cluster {
    concat::fragment { "${name} ${cluster} primary resource":
      target  => "/etc/drbd.d/${name}.res",
      content => template('drbd/resource.res.erb'),
      order   => '10',
    }
  } elsif $cluster {
    concat::fragment { "${name} ${cluster} secondary resource":
      target  => "/etc/drbd.d/${name}.res",
      content => template('drbd/resource.res.erb'),
      order   => '20',
    }
  } elsif $host1 and $ip1 and $host2 and $ip2 {
    concat::fragment { "${name} static primary resource":
      target  => "/etc/drbd.d/${name}.res",
      content => template('drbd/primary-resource.res.erb'),
      order   => '10',
    }
    concat::fragment { "${name} static secondary resource":
      target  => "/etc/drbd.d/${name}.res",
      content => template('drbd/secondary-resource.res.erb'),
      order   => '20',
    }
  } elsif $res1 and $ip1 and $res2 and $ip2 {
    concat::fragment { "${name} static stacked primary resource":
      target  => "/etc/drbd.d/${name}.res",
      content => template('drbd/primary-stacked-resource.res.erb'),
      order   => '10',
    }
    concat::fragment { "${name} static stacked secondary resource":
      target  => "/etc/drbd.d/${name}.res",
      content => template('drbd/secondary-stacked-resource.res.erb'),
      order   => '20',
    }
  } else {
    fail('Must provide either cluster, host1/host2/ip1/ip2 or res1/res2/ip1/ip2 parameters')
  }

  concat::fragment { "${name} drbd footer":
    target  => "/etc/drbd.d/${name}.res",
    content => "}\n",
    order   => '99',
  }

#  if $cluster {
#    # Import cluster nodes
#    Concat::Fragment <<| title == "${name} ${cluster} primary resource" |>>
#    Concat::Fragment <<| title == "${name} ${cluster} secondary resource" |>>
#  }

  # Due to a bug in puppet, defined() conditionals must be in a defined
  # resource to be evaluated *after* the collector instead of before.
  $_cluster = $cluster ? {
    undef   => 'static',
    default => $cluster,
  }
  drbd::resource::enable { $name:
    manage        => $manage,
    disk          => $disk,
    fs_type       => $fs_type,
    mkfs_opts     => $mkfs_opts,
    device        => $device,
    ha_primary    => $ha_primary,
    initial_setup => $initial_setup,
    cluster       => $_cluster,
    mountpoint    => $mountpoint,
    automount     => $automount,
  }
}
