define drbd::resource::up (
  $disk,
  $ha_primary,
  $initial_setup,
  $fs_type,
  $mkfs_opts,
  $device,
  $mountpoint,
  $automount,
) {

  # Ensure disk by-path link exists
  exec { "test disk by-path for ${name}":
    command => "udevadm settle",
    unless => "test -e ${disk}",
    before => Service['drbd'],
    require => [
        Exec['modprobe drbd']
      ],
  }

  # Ensure DRBD metadata is clear
  exec { "apply-al DRBD metadata for ${name}":
    command => "drbdadm apply-al ${name}",
    onlyif => "drbdadm dump-md ${name} 2>&1 | egrep 'please apply-al'",
    before => Service['drbd'],
    require => [
        Exec['modprobe drbd']
      ],
  }

  # create metadata on device, except if resource seems already initalized.
  # drbd is very tenacious about asking for aproval if there is data on the
  # volume already.
  exec { "initialize DRBD metadata for ${name}":
    command => "yes yes | drbdadm create-md ${name} -W--peer-max-bio-size=128k",
    onlyif  => "test -e ${disk} && ! drbdadm dstate ${name}",
    unless  => "drbdadm dump-md ${name} || (drbdadm cstate ${name} | egrep -q '^(PausedSync|Sync|Connected|WFConnection|StandAlone|Verify)') || (drbdadm show-gi ${name} | grep 'meta-data: need apply-al')",
    before  => Service['drbd'],
    require => [
      Exec['modprobe drbd'],
      Exec["test disk by-path for ${name}"],
      Concat["/etc/drbd.d/${name}.res"],
      Exec["apply-al DRBD metadata for ${name}"]
      ],
    notify  => Service['drbd'],
  }

  exec { "enable DRBD resource ${name}":
    command => "drbdadm up ${name}",
    onlyif  => "drbdadm dstate ${name} | egrep -q '^(Diskless/|Unconfigured|Consistent)'",
    unless  => "drbdadm show-gi ${name} | grep 'meta-data: need apply-al'",
    before  => Service['drbd'],
    require => [
      Exec["initialize DRBD metadata for ${name}"],
      Exec['modprobe drbd']
      ],
    notify  => Service['drbd'],
  }

  exec { "reuse existing DRBD resource ${name}":
    command => "drbdadm adjust ${name}",
    onlyif  => "test -e ${disk} && (drbdadm show-gi ${name} | grep 'meta-data: need apply-al')",
    before  => Service['drbd'],
    require => [
      Exec['modprobe drbd'],
      Concat["/etc/drbd.d/${name}.res"],
    ],
    notify  => Service['drbd'],
  }

  # these resources should only be applied if we are configuring the
  # primary node in our HA setup
  if $ha_primary {
    $cmd =  "/bin/true # comment to satisfy puppet syntax requirements
set -ex
if ! drbdadm primary ${name} ; then
  for i in {1..10}; do
    if drbdadm primary ${name} --force ; then
      exit 0
    fi
    sleep 0.1
  done
  drbdadm primary ${name} --force
fi
"
    # these things should only be done on the primary during initial setup
    if $initial_setup {
      exec { "drbd_make_primary_${name}":
        command => $cmd,
        unless  => "drbdadm role ${name} | egrep '^Primary'",
        notify  => Exec["drbd_format_volume_${name}"],
        before  => Exec["drbd_make_primary_again_${name}"],
        require => Service['drbd'],
      }
      $before = $automount ? {
        false   => undef,
        default => Mount[$mountpoint],
      }
      exec { "drbd_format_volume_${name}":
        command     => "mkfs.${fs_type} ${mkfs_opts} ${device}",
        timeout     => 0,
        refreshonly => true,
        require     => Exec["drbd_make_primary_${name}"],
        before      => $before,
      }
    }

    exec { "drbd_make_primary_again_${name}":
      command => $cmd,
      unless  => "drbdadm role ${name} | egrep '^Primary'",
      require => Service['drbd'],
    }

    if $automount {
      # ensure that the device is mounted
      mount { $mountpoint:
        ensure  => mounted,
        atboot  => yes,
        device  => $device,
        fstype  => 'auto',
        options => 'defaults,noauto',
        require => [
          Exec["drbd_make_primary_again_${name}"],
          File[$mountpoint],
          Service['drbd']
        ],
      }
    }
  }
}
